/*
 * Copyright (C) The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package multiplebarcode.multiplebarcode

import android.Manifest
import android.annotation.SuppressLint
import android.app.Activity
import android.app.AlertDialog
import android.app.Dialog
import android.content.Context
import android.content.DialogInterface
import android.content.Intent
import android.content.IntentFilter
import android.content.pm.PackageManager
import android.graphics.Color
import android.graphics.drawable.ColorDrawable
import android.hardware.Camera
import android.os.Build
import android.os.Bundle
import android.util.Log
import android.view.Menu
import android.view.MenuItem
import android.view.View
import android.view.Window
import android.view.WindowManager
import android.widget.Toast
import androidx.appcompat.app.AppCompatActivity
import androidx.core.app.ActivityCompat

import com.google.android.gms.common.ConnectionResult
import com.google.android.gms.common.GoogleApiAvailability
import com.google.android.gms.common.api.CommonStatusCodes
import multiplebarcode.multiplebarcode.ui.camera.CameraSource
import multiplebarcode.multiplebarcode.ui.camera.CameraSourcePreview
import multiplebarcode.multiplebarcode.ui.camera.GraphicOverlay
import com.google.android.gms.vision.MultiProcessor
import com.google.android.gms.vision.barcode.Barcode
import com.google.android.gms.vision.barcode.BarcodeDetector

import java.io.IOException
import java.util.ArrayList
import java.util.Timer
import java.util.TimerTask

/**
 * Activity for the multi-tracker app.  This app detects barcodes and displays the value with the
 * rear facing camera. During detection overlay graphics are drawn to indicate the position,
 * size, and ID of each barcode.
 */
class BarcodeCaptureActivity : AppCompatActivity(), BarcodeGraphicTracker.BarcodeUpdateListener {
    private var mCameraSource: CameraSource? = null
    private var mPreview: CameraSourcePreview? = null
    private var mGraphicOverlay: GraphicOverlay<BarcodeGraphic>? = null
    internal var useFlash = false

    internal var barcodeTimeout = 3

    internal var barcodeListStr = ArrayList<String>()

    // helper objects for detecting taps and pinches.
    //    private ScaleGestureDetector scaleGestureDetector;
    //    private GestureDetector gestureDetector;

    /**
     * Initializes the UI and creates the detector pipeline.
     */
    public override fun onCreate(icicle: Bundle?) {
        super.onCreate(icicle)
        setContentView(R.layout.barcode_capture)

        var parimaryColor = "#000000"
        var parimaryDarkColor = "#000000"

        if (intent != null && intent.extras != null) {
            barcodeTimeout = intent.getIntExtra("MULTIPLE_BARCODE_INTERVAL_SEC", barcodeTimeout)
            parimaryColor = if (intent.getStringExtra("PRIMARY_COLOR") != null) "#" + intent.getStringExtra("PRIMARY_COLOR") else parimaryColor
            parimaryDarkColor = if (intent.getStringExtra("PRIMARY_DARK_COLOR") != null) "#" + intent.getStringExtra("PRIMARY_DARK_COLOR") else parimaryDarkColor

        }

        if (supportActionBar != null) {
            supportActionBar!!.setBackgroundDrawable(ColorDrawable(Color.parseColor(parimaryColor)))
            supportActionBar!!.setDisplayShowTitleEnabled(false)
            supportActionBar!!.setDisplayShowTitleEnabled(true)
            supportActionBar!!.setDisplayHomeAsUpEnabled(true)
        }

        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.LOLLIPOP) {
            val window = window
            // clear FLAG_TRANSLUCENT_STATUS flag:
            window.clearFlags(WindowManager.LayoutParams.FLAG_TRANSLUCENT_STATUS)
            // add FLAG_DRAWS_SYSTEM_BAR_BACKGROUNDS flag to the window
            window.addFlags(WindowManager.LayoutParams.FLAG_DRAWS_SYSTEM_BAR_BACKGROUNDS)
            // finally change the color
            window.statusBarColor = Color.parseColor(parimaryDarkColor)
        }

        mPreview = findViewById<View>(R.id.preview) as CameraSourcePreview
        mGraphicOverlay = findViewById<View>(R.id.graphicOverlay) as GraphicOverlay<BarcodeGraphic>

        Timer().scheduleAtFixedRate(object : TimerTask() {
            override fun run() {
                var opacity = mGraphicOverlay!!.opacity
                opacity = opacity + 20
                if (opacity >= 255)
                    opacity = 25

                mGraphicOverlay!!.opacity = opacity
                mGraphicOverlay!!.invalidate()
            }
        }, 70, 70)


        // read parameters from the intent used to launch the activity.
        val autoFocus = intent.getBooleanExtra(AutoFocus, true)

        // Check for the camera permission before accessing the camera.  If the
        // permission is not granted yet, request permission.
        val rc = ActivityCompat.checkSelfPermission(this, Manifest.permission.CAMERA)
        if (rc == PackageManager.PERMISSION_GRANTED) {
            createCameraSource(autoFocus, useFlash)
        } else {
            requestCameraPermission()
        }


    }

    override fun onCreateOptionsMenu(menu: Menu): Boolean {
        menuInflater.inflate(R.menu.main_menu, menu)
        return super.onCreateOptionsMenu(menu)
    }

    override fun onOptionsItemSelected(item: MenuItem): Boolean {
        if (item.itemId == R.id.menu_flash) {
            useFlash = !useFlash
            mCameraSource!!.setFlashMode(if (useFlash) Camera.Parameters.FLASH_MODE_TORCH else Camera.Parameters.FLASH_MODE_OFF)
            if (useFlash)
                item.setIcon(R.drawable.ic_flash_off)
            else
                item.setIcon(R.drawable.ic_flash_on)
        } else if (item.itemId == android.R.id.home) {
            val data = Intent()
            data.putExtra(BarcodeObject, barcodeListStr)
            setResult(Activity.RESULT_OK, data)
            finish()
        }
        return super.onOptionsItemSelected(item)
    }

    /**
     * Handles the requesting of the camera permission.  This includes
     * showing a "Snackbar" message of why the permission is needed then
     * sending the request.
     */
    private fun requestCameraPermission() {
        Log.w(TAG, "Camera permission is not granted. Requesting permission")

        val permissions = arrayOf(Manifest.permission.CAMERA)

        if (!ActivityCompat.shouldShowRequestPermissionRationale(this,
                        Manifest.permission.CAMERA)) {
            ActivityCompat.requestPermissions(this, permissions, RC_HANDLE_CAMERA_PERM)
            return
        }

    }

    @SuppressLint("InlinedApi")
    private fun createCameraSource(autoFocus: Boolean, useFlash: Boolean) {
        val context = applicationContext

        val barcodeDetector = BarcodeDetector.Builder(context).build()
        val barcodeFactory = BarcodeTrackerFactory(mGraphicOverlay!!, this)
        barcodeDetector.setProcessor(
                MultiProcessor.Builder(barcodeFactory).build())
        if (!barcodeDetector.isOperational) {
            Log.w(TAG, "Detector dependencies are not yet available.")
            // Check for low storage.  If there is low storage, the native library will not be
            // downloaded, so detection will not become operational.
            val lowstorageFilter = IntentFilter(Intent.ACTION_DEVICE_STORAGE_LOW)
            val hasLowStorage = registerReceiver(null, lowstorageFilter) != null

            if (hasLowStorage) {
                Toast.makeText(this, R.string.low_storage_error, Toast.LENGTH_LONG).show()
                Log.w(TAG, getString(R.string.low_storage_error))
            }
        }

        var builder: CameraSource.Builder = CameraSource.Builder(applicationContext, barcodeDetector)
                .setFacing(CameraSource.CAMERA_FACING_BACK)
                .setRequestedPreviewSize(1600, 1024)
                .setRequestedFps(15.0f)

        // make sure that auto focus is an available option
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.ICE_CREAM_SANDWICH) {
            builder = builder.setFocusMode(
                    if (autoFocus) Camera.Parameters.FOCUS_MODE_CONTINUOUS_PICTURE else null)
        }

        mCameraSource = builder
                .setFlashMode(if (useFlash) Camera.Parameters.FLASH_MODE_TORCH else null)
                .build()
    }


    override fun onResume() {
        super.onResume()
        startCameraSource()
    }


    override fun onPause() {
        super.onPause()
        if (mPreview != null) {
            mPreview!!.stop()
        }
    }

    override fun onDestroy() {
        super.onDestroy()
        if (mPreview != null) {
            mPreview!!.release()
        }
    }


    override fun onRequestPermissionsResult(requestCode: Int,
                                            permissions: Array<String>,
                                            grantResults: IntArray) {
        if (requestCode != RC_HANDLE_CAMERA_PERM) {
            Log.d(TAG, "Got unexpected permission result: $requestCode")
            super.onRequestPermissionsResult(requestCode, permissions, grantResults)
            return
        }

        if (grantResults.size != 0 && grantResults[0] == PackageManager.PERMISSION_GRANTED) {
            Log.d(TAG, "Camera permission granted - initialize the camera source")
            // we have permission, so create the camerasource
            val autoFocus = intent.getBooleanExtra(AutoFocus, false)
            val useFlash = intent.getBooleanExtra(UseFlash, false)
            createCameraSource(autoFocus, useFlash)
            return
        }

        Log.e(TAG, "Permission not granted: results len = " + grantResults.size +
                " Result code = " + if (grantResults.size > 0) grantResults[0] else "(empty)")

        val listener = DialogInterface.OnClickListener { dialog, id ->
            val data = Intent()
            data.putExtra(BarcodeObject, barcodeListStr)
            setResult(Activity.RESULT_OK, data)
            finish()
        }

        val builder = AlertDialog.Builder(this)
        builder.setTitle(null)
                .setMessage(R.string.no_camera_permission)
                .setPositiveButton(R.string.ok, listener)
                .show()
    }

    /**
     * Starts or restarts the camera source, if it exists.  If the camera source doesn't exist yet
     * (e.g., because onResume was called before the camera source was created), this will be called
     * again when the camera source is created.
     */
    @Throws(SecurityException::class)
    private fun startCameraSource() {
        // check that the device has play services available.
        val code = GoogleApiAvailability.getInstance().isGooglePlayServicesAvailable(
                applicationContext)
        if (code != ConnectionResult.SUCCESS) {
            val dlg = GoogleApiAvailability.getInstance().getErrorDialog(this, code, RC_HANDLE_GMS)
            dlg.show()
        }

        if (mCameraSource != null) {
            try {
                mGraphicOverlay?.let { mPreview!!.start(mCameraSource!!, it) }
            } catch (e: IOException) {
                Log.e(TAG, "Unable to start camera source.", e)
                mCameraSource!!.release()
                mCameraSource = null
            }

        }
    }


    override fun onBarcodeDetected(barcode: Barcode?) {
        //do something with barcode data returned
        if (barcode != null) {
            if (!barcodeListStr.contains(barcode.rawValue)) {
                if (barcodeListStr.isEmpty()) {
                    Timer().schedule(object : TimerTask() {
                        override fun run() {
                            val data = Intent()
                            data.putExtra(BarcodeObject, barcodeListStr)
                            setResult(Activity.RESULT_OK, data)
                            finish()
                        }
                    }, (barcodeTimeout * 1000).toLong())
                }
                barcodeListStr.add(barcode.rawValue)
                //            int sx = barcode.cornerPoints[0].x;
                //            int ex = barcode.cornerPoints[1].x;
                //            int sy = barcode.cornerPoints[0].y;
                //            int ey = barcode.cornerPoints[1].y;
                ////            Log.e("COMING", "SX -" + sx + " EX -" + ex + " SY -" + sy + " EY -" + ey);
                //
                //            int startXPoint = mGraphicOverlay.width / 2 - mGraphicOverlay.squareSize;
                //            int endXPoint = mGraphicOverlay.width / 2 + mGraphicOverlay.squareSize;
                //            int startYPoint = mGraphicOverlay.height / 2 - mGraphicOverlay.squareSize;
                //            int endYPoint = mGraphicOverlay.height / 2 + mGraphicOverlay.squareSize;
                ////            Log.e("STATIC", "SX -" + startXPoint + " EX -" + endXPoint + " SY -" + startYPoint + " EY -" + endYPoint);
                //
                //            if (sx >= startXPoint && sx <= endXPoint &&
                //                    ex >= startXPoint && ex <= endXPoint &&
                //                    sy >= startYPoint && sy <= endYPoint &&
                //                    ey >= startYPoint && ex <= endYPoint) {
                //                if (barcodeListStr.isEmpty()) {
                //                    new Timer().schedule(new TimerTask() {
                //                        @Override
                //                        public void run() {
                //                            Intent data = new Intent();
                //                            data.putExtra(BarcodeObject, barcodeListStr);
                //                            setResult(RESULT_OK, data);
                //                            finish();
                //                        }
                //                    }, barcodeTimeout * 1000);
                //                }
                //                barcodeListStr.add(barcode.rawValue);
                //            }
            }
        }
    }

    companion object {
        private val TAG = "Barcode-reader"

        // intent request code to handle updating play services if needed.
        private val RC_HANDLE_GMS = 9001

        // permission request codes need to be < 256
        private val RC_HANDLE_CAMERA_PERM = 2

        // constants used to pass extra data in the intent
        val AutoFocus = "AutoFocus"
        val UseFlash = "UseFlash"
        val BarcodeObject = "Barcode"
    }
}
